/******************************************************************************\
* $Id:$
*
* Copyright(c) 2015 SAS Institute Inc., Cary, NC, USA. All Rights Reserved.
*
* Name: checkXPTFiles.sas
*
* Purpose: check whether submission .xpt files conform to technical specs
*            - only one data set in each .xpt file 
*            - data set name/XPT file name are the same
*
* Author: Ben Bocchicchio, Frank Roediger
*
* Support: SAS(r) Solutions OnDemand
*
* Input: 
*
* Output: N/A
*
* Parameters: N/A
*
* Dependencies/Assumptions: 
*
* Usage: -- use %LET to assign path of the module root folder
*        -- use %LET to assign length of the longest overall path
*
* History:
* ddmmmyyyy userid         description                             (Change Code)
* 27FEB2015 FrankRoediger  initial development
* 16APR2015 FrankRoediger  use DATA _null_ / INFILE to read v5 transport files
* 02MAY2015 FrankRoediger  verify contents are V5 transport file
\******************************************************************************/

%*******************************************************************************
%* housekeeping: assign macro parameters
%******************************************************************************;
%let moduleRoot=C:\PharmaSUG2015\Folders;
%let logicRecLen=500;


%*******************************************************************************
%* define the principal macro
%******************************************************************************;
%macro checkXPTFiles;
   options noxwait dtreset pageno=1;

   %****************************************************************************
   %* housekeeping: delete WORK.ALLTEST
   %***************************************************************************;
   proc datasets library=work nolist nowarn;
      delete allTest;
   quit;


   %****************************************************************************
   %* create an inventory of .xpt files in the submission
   %****************************************************************************
   %* execute DIR to create a .txt file with all the sub-folders of module root
   %***************************************************************************;
   %sysexec dir "&moduleRoot\*.xpt"  /s > "&moduleRoot\dir_chkXPTFiles.txt";

   data files
         (keep=allPath
               fileName)
       ;
      length fileName   $ 100
             allPath    $ &logicRecLen
            ;

      retain allPath;

      infile "&moduleRoot\dir_chkXPTFiles.txt"
         lrecl=&logicRecLen
         pad
         end=eof;

      input @ 1 inRec $char&logicRecLen..;

      if substr(inRec,1,14) eq ' Directory of ' then do;
         allPath=strip(substr(inRec,15));
      end;
      else if substr(inRec,3,1) eq '/' and
              substr(inRec,6,1) eq '/' and
              not index(inRec,'<DIR>')      then do;
         fileName=strip(substr(inRec,40));
         output files;
      end;

   run;

   proc sort data=files
             out =paths
             nodupkey;
      by allPath;
   run;


   %****************************************************************************
   %* create a macro variable for each .xpt transport file
   %***************************************************************************;
   data _null_;
      set files end=eof;

      call symput('path_' || strip(put(_n_,4.))
                 ,strip(allPath) || '\' || strip(fileName)
                 );

      call symput('xptName_' || strip(put(_n_,4.))
                 ,strip(fileName)
                 );

      if eof then 
         call symput('pathCnt',strip(put(_n_,2.)));
   run;


   %do idx=1 %to &pathCnt;
      %*************************************************************************
      %* NOTE: LIBNAME statements with XPORT engine can be extremely inefficient
      %*       when a .xpt file is very big.  the code is commented out but left
      %*       in the program for future reference
      %*
      %* NOTE: DICTIONARY views or PROC Contents can be used to retrieve
      %*       metadata about XPORT engine libraries
      %************************************************************************;
      %* libname x1 xport "&&path_&idx";
      %*
      %* proc contents data=x1._all_
      %*               out =xptMetaData;
      %* run;
      %*
      %* proc sql;
      %*    create table x1Tables as
      %*       select *
      %*       from dictionary.Tables
      %*          where libname eq 'X1';
      %* quit;
      %*
      %* libname x1 clear;
 
      
      data scanResults
            (keep=memCnt
                  matchSw
                  v5Sw
                  memname
            );

         length memName     $ 32
                xptDsetName $ 45
               ;

         %**********************************************************************
         %* because v5 transport files use 80-byte logical records, LRECL should 
         %*    be a multiple of 80.  The most efficient I/O would be the largest
         %*    acceptable value (32640 -- largest even multiple of 80 less than
         %*    or equal to 32676), but SAS string functions do not work 
         %*    efficiently on character fields that long.  benchmark tests
         %*    indicated that 3200 was a good compromise between I/O efficiency
         %*    and string function efficiency
         %* when RECFM=N, the UNBUFFERED option is set by default.  the END=
         %*    option is not available with UNBUFFERED, so we use the EOF= 
         %*    option instead
         %* as soon as one of the 3 rules is failed, the program stops reading
         %*    the XPT file
         %*********************************************************************;
         infile "&&path_&idx"
            lrecl=3200
            recfm=n
            eof=endRead;

         retain v5Sw
                prevMemCnt  0
                xptDsetName
                memName
               ;

         %**********************************************************************
         %* INREC is $char3200. so that it will accommodate the entire contents
         %*    of the input buffer
         %*********************************************************************;
         input inRec $char3200.;

         %**********************************************************************
         %* xptDsetName contains a string value with the way data set names are
         %*    represented in V5 transport files
         %**********************************************************************
         %* NOTE: 1st record in V5 transport files contain 
         %*          LIBRARY HEADER RECORD!
         %* other transport files (created with PROC CPORT) do not have this
         %*********************************************************************;
         if _n_ eq 1 then do;
            xptDsetName='SAS     ' 
                     || put(upcase(scan("&&xptName_&idx",1,'.')),$char8.)
                     || 'SASDATA';
            put xptDsetName=;
            memName=scan("&&xptName_&idx",1,'.');
            put memName=;
            if index(inRec,'LIBRARY HEADER RECORD!') then  
               v5Sw=1;
            else do;
               v5Sw=0;
               link endRead;
            end;
         end;

         %**********************************************************************
         %* *DSCRPTR HEADER RECORD!!!!!!! indicates the start of a new data set
         %*********************************************************************;
         memCnt+count(inRec,'*DSCRPTR HEADER RECORD!!!!!!!');
         if memCnt gt 1 then
            link endRead;

         %**********************************************************************
         %* if xptDsetName is found, the V5 transport file name matches an  
         %*    embedded data set name 
         %*********************************************************************;
         if index(inRec,strip(xptDsetName)) then 
            matchSw+1;

         return;

      endRead:; 
         output;
         stop;
       
      run;

      data test;
         length testResult $ 4
                message    $ 150
                testSeq      8
                testID       8
               ;

         set scanResults;

         %**********************************************************************
         %* create message line #1 -- Summary
         %*********************************************************************;
         testSeq   =1+(&idx/1000)+.0001;
         xptID     =round(testSeq,.001);

         if v5Sw eq 0 then do;
            testResult='FAIL';
            message   ='XPT is not a V5 transport file';
            output;
         end;
         else if memCnt gt 1 then do;
            testResult='FAIL';
            message   ='XPT contains more than 1 data set';
            output;
         end;
         else if matchSw ne 1 then do;
            testResult='FAIL';
            message   ='XPT file name does not match data set name';
            output;
         end;
         else do;
            testResult='Pass';
            message   ='XPT transport file meets Conformance Guide Rules';
            output;
         end;

         %**********************************************************************
         %* create message line #2 -- detail
         %*********************************************************************;
         testResult=' ';
         message   ='[' 
                 || upcase(scan("&&xptName_&idx",1,'.')) 
                 || '] : '
                 || strip("&&path_&idx")
                ;
         testSeq   =1+(&idx/1000)+.0002;
         xptID     =round(testSeq,.001);
         output;

         keep testResult
              message
              testSeq
              xptID
             ;

      run;
      %*************************************************************************
      %* append the current results into the data set with all other results 
      %************************************************************************;
      proc append base=allTest
                  data=test;
      run;

   %end;
   
   ods html close;
   ods listing;
   options ls=150 ps=5000 pageno=1;

   proc sort data=allTest;
      by xptID;
   run;

   proc print data=allTest noobs;
      by xptID;
      id xptID;
      var testResult
          message;
   run;

   %****************************************************************************
   %* housekeeping: delete dir_chkSubFld.txt 
   %***************************************************************************;
   %sysexec del "&moduleRoot\dir_chkXPTFiles.txt";

%mend checkXPTFiles;

%checkXPTFiles;

